# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.
from __future__ import annotations
import numpy as np
import typing as t_
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateAnalysisResults
from pwspy.dataTypes import Acquisition, FluorescenceImage
from enum import Enum
from pwspy.analysis.pws import PWSAnalysisResults
from pwspy.analysis.dynamics import DynamicsAnalysisResults


class _AnalysisTypes(Enum):
    """An Enumeration of the possible analysis types."""
    PWS = "pws"
    DYN = "dyn"


class _PlotFields(Enum):
    """An enumerator of the possible images that can be displayed.
    The first item of each tuple indicates which analysis must be available for the image to be displayed.
    The second item of each tuple is a string value matching the attribute name of the associated analysisResults class."""
    Thumbnail = (None, 'thumbnail')
    Fluorescence0 = (None, 'fluorescence0')
    Fluorescence1 = (None, 'fluorescence1')
    Fluorescence2 = (None, 'fluorescence2')
    Fluorescence3 = (None, 'fluorescence3')
    Fluorescence4 = (None, 'fluorescence4')
    Fluorescence5 = (None, 'fluorescence5')
    Fluorescence6 = (None, 'fluorescence6')
    Fluorescence7 = (None, 'fluorescence7')
    Fluorescence8 = (None, 'fluorescence8')
    Fluorescence9 = (None, 'fluorescence9')  # Hopefully we never need nearly this many.
    # PWS specific
    OpdPeak = (_AnalysisTypes.PWS, 'opdPeak')  # Even though this isn't really a name of an analysis field we need something unique for this enum value.
    SingleWavelength = (_AnalysisTypes.PWS, 'singleWavelength')
    MeanReflectance = (_AnalysisTypes.PWS, 'meanReflectance')
    RMS = (_AnalysisTypes.PWS, 'rms')
    AutoCorrelationSlope = (_AnalysisTypes.PWS, 'autoCorrelationSlope')
    RSquared = (_AnalysisTypes.PWS, 'rSquared')
    Ld = (_AnalysisTypes.PWS, 'ld')
    # Dynamics specific
    RMS_t_squared = (_AnalysisTypes.DYN, 'rms_t_squared')
    Diffusion = (_AnalysisTypes.DYN, 'diffusion')
    DynamicsReflectance = (_AnalysisTypes.DYN, 'meanReflectance')


_FluorescencePlotFields = [_PlotFields.Fluorescence0, _PlotFields.Fluorescence1, _PlotFields.Fluorescence2,
                           _PlotFields.Fluorescence3, _PlotFields.Fluorescence4, _PlotFields.Fluorescence5,
                           _PlotFields.Fluorescence6, _PlotFields.Fluorescence7, _PlotFields.Fluorescence8,
                           _PlotFields.Fluorescence9, ]


class AnalysisPlotter:
    PlotFields = _PlotFields
    _fluorescencePlotFields = _FluorescencePlotFields

    AnalysisResultsComboType = t_.Union[
        ConglomerateAnalysisResults,  # Internally we will use the ConglomerateAnalysisResults but we accept a regular tuple as well.
        t_.Tuple[
            t_.Optional[PWSAnalysisResults],
            t_.Optional[DynamicsAnalysisResults]
        ]]

    def __init__(self, acq: Acquisition, analysis: AnalysisResultsComboType = None, initialField=PlotFields.Thumbnail):
        self._analysisField: AnalysisPlotter.PlotFields = initialField
        self._analysis: ConglomerateAnalysisResults = ConglomerateAnalysisResults(*analysis)
        self._acq: Acquisition = acq
        self._data: np.ndarray = None

    @property
    def analysis(self) -> ConglomerateAnalysisResults:
        return self._analysis

    @property
    def acq(self) -> Acquisition:
        return self._acq

    @property
    def data(self) -> np.ndarray:
        return self._data

    @property
    def analysisField(self) -> AnalysisPlotter.PlotFields:
        return self._analysisField

    def changeData(self, field: _PlotFields):
        assert isinstance(field, AnalysisPlotter.PlotFields)
        self._analysisField = field
        if field is _PlotFields.Thumbnail:  # Load the thumbnail from the ICMetadata object
            self._data = self._acq.getThumbnail()
        elif field in _FluorescencePlotFields:  # Open the fluorescence image.
            idx = _FluorescencePlotFields.index(field)  # Get the number for the fluorescence image that has been selected.
            self._data = FluorescenceImage.fromMetadata(self._acq.fluorescence[idx]).data
        else:
            anType, paramName = field.value
            if anType == _AnalysisTypes.PWS:
                analysis = self._analysis.pws
            elif anType == _AnalysisTypes.DYN:
                analysis = self._analysis.dyn
            else:
                raise TypeError("Unidentified analysis type")
            if analysis is None:
                raise ValueError(f"Analysis Plotter for {self._acq.filePath} does not have an analysis file.")
            if field is _PlotFields.OpdPeak:  # Return the index corresponding to the max of that pixel's opd funtion.
                opd, opdIndex = self._analysis.pws.opd
                self._data = opdIndex[np.argmax(opd, axis=2)]
            elif field is _PlotFields.SingleWavelength:  # Return the image of the middle wavelength reflectance.
                _ = self._analysis.pws.reflectance.data
                self._data = _[:, :, _.shape[2]//2] # + self.analysis.pws.meanReflectance # It actually looks better without the meanReflectance added
            else:
                self._data = getattr(analysis, paramName)
        assert len(self._data.shape) == 2

    def setMetadata(self, md: Acquisition, analysis: t_.Optional[AnalysisResultsComboType] = None):
        self._analysis = ConglomerateAnalysisResults(*analysis) if not isinstance(analysis, ConglomerateAnalysisResults) else analysis # In case this was a regular tuple, convert to our convenience class
        self._acq = md
        self.changeData(self._analysisField)




# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

# -*- coding: utf-8 -*-
"""
This file is used to install the pwspy package. for example navigate in your terminal to the directory containing this
file and type `pip install .`. This file is also used by the Conda recipe (buildscripts/conda)
"""
import pathlib
from setuptools import setup, find_packages
import setuptools_scm

HERE = pathlib.Path(__file__).parent  # The directory containing this file
README = (HERE / "README.md").read_text()  # The text of the README file

setup(name='pwspy_gui',
      version=setuptools_scm.get_version(write_to="src/pwspy_gui/version.py"),
      description='A collection of GUIs providing users with easy access to PWS data analysis.',
      long_description=README,
      long_description_content_type="text/markdown",
      author='Nick Anthony',
      author_email='nicholas.anthony@northwestern.edu',
      url='https://github.com/nanthony21/pwspy_gui',
      python_requires='>=3.7',
      install_requires=['numpy',
                        'matplotlib',
                        'psutil',
                        'shapely',
                        'pandas',
                        'jsonschema',
                        'google-api-python-client',
                        'google-auth-httplib2',
                        'google-auth-oauthlib',
                        'PyQt5',
                        'pwspy>=0.2.4',  # Core pws package, available on backmanlab anaconda cloud account.
                        'mpl_qt_viz>=1.0.7',  # Plotting package available on PyPi and the backmanlab anaconda cloud account. Written for this project by Nick Anthony
                        'descartes',
                        'cachetools>=4'],
      package_dir={'': 'src'},
      package_data={'pwspy_gui': ['_resources/*',
                              'PWSAnalysisApp/_resources/*']},
      packages=find_packages('src'),
	  entry_points={'gui_scripts': [
          'PWSAnalysis = pwspy_gui.PWSAnalysisApp.__main__:main',
          "ERCreator = pwspy_gui.ExtraReflectanceCreator.__main__:main"
      ]}
	)



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from __future__ import annotations
from PyQt5.QtWidgets import QComboBox, QWidget, QVBoxLayout
from typing import Optional
import typing as t_
import pwspy.dataTypes as pwsdt
from pwspy_gui.PWSAnalysisApp.componentInterfaces import ROIManager
from pwspy_gui.PWSAnalysisApp.utilities.conglomeratedAnalysis import ConglomerateAnalysisResults

from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._widgets import AnalysisPlotter
from pwspy_gui.PWSAnalysisApp.sharedWidgets.plotting._roiPlot import RoiPlot


class AnalysisViewer(AnalysisPlotter, QWidget):
    """This class is a window that provides convenient viewing of a pws acquisition, analysis, and related images.
    It expands upon the functionality of `BigPlot` which handles ROIs but not analysis images."""
    def __init__(self, metadata: pwsdt.Acquisition, analysisLoader: AnalysisPlotter.AnalysisResultsComboType, title: str, roiManager: ROIManager, parent=None,
                 initialField=AnalysisPlotter.PlotFields.Thumbnail, flags=None):
        if flags is not None:
            QWidget.__init__(self, parent=parent, flags=flags)
        else:
            QWidget.__init__(self, parent=parent)
        AnalysisPlotter.__init__(self, metadata, analysisLoader, initialField=initialField)

        self.roiPlot = RoiPlot(metadata, metadata.getThumbnail(), roiManager=roiManager, parent=self)

        self.setWindowTitle(title)
        self.analysisCombo = QComboBox(self)
        self._populateFields()

        self.roiPlot.layout().itemAt(0).insertWidget(0, self.analysisCombo)  # This is sketchy, oh well.

        l = QVBoxLayout()
        l.setContentsMargins(0, 0, 0, 0)
        l.addWidget(self.roiPlot)
        self.setLayout(l)

        self.changeData(self.analysisField)

    def _populateFields(self):
        currField = self.analysisCombo.currentText()
        try:
            self.analysisCombo.disconnect()
        except:
            pass #Sometimes there is nothing to disconnect
        self.analysisCombo.clear()
        _ = self.PlotFields  # Just doing this to make for less typing later.
        items = [_.Thumbnail]
        for i in [_.MeanReflectance, _.RMS, _.AutoCorrelationSlope, _.RSquared, _.Ld]:
            try:
                if hasattr(self.analysis.pws, i.value[1]):  # This will raise a key error if the analysis object exists but the requested item is not found
                    items.append(i)
            except KeyError:
                pass
        for i in [_.RMS_t_squared, _.Diffusion, _.DynamicsReflectance]:
            try:
                if hasattr(self.analysis.dyn, i.value[1]):  # This will raise a key error if the analysis object exists but the requested item is not found
                    items.append(i)
            except KeyError:
                pass
        if self.analysis.pws is not None:
            if 'reflectance' in self.analysis.pws.file.keys(): #This is the normalized 3d data cube. needed to generate the opd.
                items.append(_.OpdPeak)
                items.append(_.SingleWavelength)
        for i in range(len(self.acq.fluorescence)):
            items.append(self._fluorescencePlotFields[i])
        self.analysisCombo.addItems([i.name for i in items])
        if currField in [i.name for i in items]:
            self.analysisCombo.setCurrentText(currField) #Maintain the same field if possible.
        self.analysisCombo.currentTextChanged.connect(self.changeDataByName)  # If this line comes before the analysisCombo.addItems line then it will get triggered when adding items.

    def changeDataByName(self, field: str):
        field = [enumField for enumField in self.PlotFields if enumField.name == field][0] #This function recieves the name of the Enum item. we want to get the enum item itself.
        self.changeData(field)

    def changeData(self, field: AnalysisPlotter.PlotFields):
        """Change which image associated with the PWS acquisition we want to view."""
        super().changeData(field)
        if self.analysisCombo.currentText() != field.name:
            self.analysisCombo.setCurrentText(field.name)
        self.roiPlot.setImageData(self.data)

    def setMetadata(self, md: Acquisition, analysis: Optional[ConglomerateAnalysisResults] = None):
        """Change this widget to display data for a different acquisition and optionally an analysis."""
        try:
            super().setMetadata(md, analysis)
        except ValueError:  # Trying to set new metadata may result in an error if the new analysis/metadata can't plot the currently set analysisField
            self.changeData(self.PlotFields.Thumbnail)  # revert back to thumbnail which should always be possible
            super().setMetadata(md, analysis)
        self.roiPlot.setMetadata(md)
        self._populateFields()



# Copyright 2018-2020 Nick Anthony, Backman Biophotonics Lab, Northwestern University
#
# This file is part of PWSpy.
#
# PWSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# PWSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with PWSpy.  If not, see <https://www.gnu.org/licenses/>.

from PyQt5.QtWidgets import QTreeWidget, QWidget, QTreeWidgetItem, QDialog, QGridLayout


class DictDisplayTree(QTreeWidget):
    def __init__(self, parent: QWidget, value: dict):
      super().__init__(parent)
      self.setHeaderHidden(True)
      self.fillItem(self.invisibleRootItem(), value)

    def fillItem(self, item, value):
      item.setExpanded(True)
      if type(value) is dict:
        for key, val in sorted(value.items()):
          child = QTreeWidgetItem()
          if isinstance(val, (tuple, list, dict)):
              child.setText(0, key)
              item.addChild(child)
              self.fillItem(child, val)
          else:
              child.setText(0, f"{key}: {str(val)}")
              item.addChild(child)
      elif isinstance(value, (tuple, list)):
        for val in value:
          child = QTreeWidgetItem()
          item.addChild(child)
          if type(val) is dict:
            child.setText(0, '[dict]')
            self.fillItem(child, val)
          elif type(val) is list:
            child.setText(0, '[list]')
            self.fillItem(child, val)
          else:
            child.setText(0, str(val))
          child.setExpanded(True)
      else:
        child = QTreeWidgetItem()
        child.setText(0, str(value))
        item.addChild(child)

class DictDisplayTreeDialog(QDialog):
    def __init__(self, parent: QWidget, value: dict, title: str = None):
        super().__init__(parent)
        if title is not None:
            self.setWindowTitle(title)
        l = QGridLayout()
        l.addWidget(DictDisplayTree(self, value))
        self.setLayout(l)



from PyQt5.QtCore import pyqtSignal, QTimer
from PyQt5.QtWidgets import QFrame, QWidget, QCheckBox, QPushButton, QVBoxLayout, QLabel
from pwspy_gui.PWSAnalysisApp.plugins.acquisitionAwareROIDrawer._control import Options
import typing as t_


class OptionsPanel(QWidget):
    animateTimerFired = pyqtSignal(int)
    optionsChanged = pyqtSignal(int, Options)

    def __init__(self, initialOptions: t_.Sequence[Options], parent: QWidget = None):
        super().__init__(parent=parent)
        l = QVBoxLayout()
        self._subPanels = []
        for i, option in enumerate(initialOptions):
            w = _SingleOptionPanel(self, option)
            w.optionsChanged.connect(lambda option, ii=i: self.optionsChanged.emit(ii, option))
            w.animateTimerFired.connect(lambda ii=i: self.animateTimerFired.emit(ii))
            self._subPanels.append(w)
            l.addWidget(w)
        l.addWidget(QWidget(parent=self), stretch=1)  # This widget just pushes the others upwards.
        self.setLayout(l)

    def getOptions(self, axis: int) -> Options:
        return self._subPanels[axis].getOptions()

    def setOptions(self, axis: int, options: Options):
        self._subPanels[axis].setOptions(options)


class _SingleOptionPanel(QFrame):
    animateTimerFired = pyqtSignal()
    optionsChanged = pyqtSignal(Options)

    def __init__(self, parent: QWidget = None, initialOptions: Options = None):
        super().__init__(parent=parent)
        self._copyTimeCB = QCheckBox("Copy ROI changes along axis", parent=self)
        self._copyTimeCB.stateChanged.connect(lambda: self.optionsChanged.emit(self.getOptions()))
        # self._trackImCB = QCheckBox("Track cell movement", parent=self)
        # self._trackImCB.stateChanged.connect(lambda: self.optionsChanged.emit(self.getOptions()))
        self._animateBtn = QPushButton("Animate axis", parent=self)
        self._animateBtn.setCheckable(True)

        self._animateBtn.toggled.connect(self._handleAnimateCheck)
        self._animateTimer = QTimer(self)
        self._animateTimer.setInterval(100)
        self._animateTimer.setSingleShot(False)
        self._animateTimer.timeout.connect(self.animateTimerFired.emit)

        l = QVBoxLayout()
        label = QLabel("Options:")
        f = label.font()
        f.setBold(True)
        label.setFont(f)
        l.addWidget(label)
        l.addWidget(self._copyTimeCB)
        # l.addWidget(self._trackImCB)
        l.addWidget(self._animateBtn)
        l.addWidget(QWidget(self), stretch=1)  # This pushes everything up to the top.
        self.setLayout(l)
        self.setFrameStyle(QFrame.Box)

        if initialOptions is not None:
            self.setOptions(initialOptions)

    def _handleAnimateCheck(self, checked: bool):
        if checked:
            self._animateTimer.start()
        else:
            self._animateTimer.stop()

    def getOptions(self) -> Options:
        return Options(
            copyAlong=self._copyTimeCB.isChecked(),
            trackMovement=False)  # self._trackImCB.isChecked())

    def setOptions(self, options: Options):
        self._copyTimeCB.setChecked(options.copyAlong)
        # self._trackImCB.setChecked(options.trackMovement)